# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

from ursa_major.handlers.base import BaseHandler
from ursa_major import MBS_BUILD_STATES
from ursa_major.logger import log
from ursa_major.utils import find_latest_module


class AddModuleHandler(BaseHandler):
    """Handler to add a module to config file and add its koji tag to inheritance data"""

    def run(self):
        self.dry_run = self.args.dry_run
        self.debug = self.args.debug

        self.tag_config_file = self.args.tag_config_file
        self.load_tag_config()

        # the koji tag we operate against
        self.tag = self.args.tag
        self.connect_mbs()
        self.connect_koji(dry_run=self.dry_run)

        module_name = self.args.module_name
        module_stream = self.args.module_stream
        module_priority = self.args.priority

        if self.args.module_requires:
            module_requires = dict(self.args.module_requires)
        else:
            module_requires = None

        if self.args.module_buildrequires:
            module_buildrequires = dict(self.args.module_buildrequires)
        else:
            module_buildrequires = None

        # find out the latest module build in MBS which matches the specified module
        match_modules = self.mbs.get_modules(
            buildrequires=module_buildrequires,
            requires=module_requires,
            name=module_name,
            stream=module_stream,
            state=MBS_BUILD_STATES['ready'])

        tag_conf = self.tag_config.get(self.tag, {})

        if match_modules:
            log.info("%d build(s) for specified module are found in MBS", len(match_modules))
            tags_from = tag_conf.get("tags_from", self.default_tags_from)
            find_latest_module(match_modules, tags_from, self.koji)
        else:
            log.warning("Can't find any built module in MBS for the specified tag")

        # all MBS modules belong to this name+stream
        name_stream_modules = self.mbs.get_modules(
              name=module_name,
              stream=module_stream,
              state=MBS_BUILD_STATES['ready'])
        name_stream_tags_in_mbs = [m['koji_tag'] for m in name_stream_modules]

        inheritance_data = self.koji.get_inheritance_data(self.tag)
        inheritance_tags = [x['name'] for x in inheritance_data]

        # tags in koji inheritance that belong to this name+stream module
        name_stream_tags_in_koji = list(set(name_stream_tags_in_mbs) & set(inheritance_tags))
        # inheritance data but exclude the tags belong to this name+stream module
        non_name_stream_inheritance = [
            x for x in inheritance_data if x['name'] not in name_stream_tags_in_koji
        ]

        # priority values in inheritance that not belong to this name+stream modules
        other_inheritance_priorities = [x['priority'] for x in non_name_stream_inheritance]

        config_modules = tag_conf.get('modules', [])
        # priority values in config file that not belong to this name+stream modules
        other_config_priorities = [
            m['priority'] for m in config_modules if not (
                m['name'] == module_name and m['stream'] == module_stream
            )
        ]

        if module_priority in other_config_priorities:
            raise ValueError("Priority {} is used by other modules in config".format(
                module_priority))
        if module_priority in other_inheritance_priorities:
            raise ValueError("Priority {} is used by other modules in tag {}'s "
                             "inheritance".format(module_priority, self.tag))

        self.add_module_to_config(self.tag, module_name, module_stream,
                                  priority=module_priority, requires=module_requires,
                                  buildrequires=module_buildrequires, dry_run=self.dry_run)

    def add_module_to_config(self, tag, name, stream, priority,
                             requires=None, buildrequires=None, dry_run=False):
        """
        Add the module config to tag config file

        :param tag: tag name, module config will be added under this tag
        :param name: module name
        :param stream: module stream
        :param priority: priority value in koji inheritance data
        :param requires: module requires
        :param buildrequires: module buildrequires
        :param dry_run: dry_run mode
        """
        tag_config = self.tag_config
        config = tag_config.setdefault(tag, {})
        modules = config.setdefault('modules', [])
        module_config = dict(name=name,
                             stream=stream,
                             priority=priority)
        if requires:
            module_config['requires'] = requires
        if buildrequires:
            module_config['buildrequires'] = buildrequires

        modules_with_same_name_stream = [m for m in modules if
                                         m['name'] == name and
                                         m['stream'] == stream]

        if len(modules_with_same_name_stream) > 1:
            raise RuntimeError("Unexpected error, found multiple modules in config "
                               "file with name '{}' and stream '{}'".format(
                                   name, stream))

        if len(modules_with_same_name_stream) == 0:
            log.info("Adding module config to tag %s in tag config file:\n%s",
                     tag, module_config)
            modules.append(module_config)

        if len(modules_with_same_name_stream) == 1:
            if module_config == modules_with_same_name_stream[0]:
                log.info("The specified module already exists, nothing to change")
                return
            log.info("Found one module in config with same name and stream")
            modules.remove(modules_with_same_name_stream[0])
            modules.append(module_config)

        log.info("Going to update tag config file with new module: {}".format(module_config))
        if dry_run:
            log.info("Dry run finished.")
        else:
            self.write_tag_config()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import logging

__logger__ = None


class CallCounted:
    """Decorator to determine number of calls for a method"""

    def __init__(self, method):
        self.method = method
        self.counter = 0

    def __call__(self, *args, **kwargs):
        self.counter += 1
        return self.method(*args, **kwargs)


def get_logger():
    global __logger__
    if __logger__ is not None:
        return __logger__
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s:%(levelname)-8s]: %(message)s')
    console.setFormatter(formatter)
    logging.getLogger(__name__).addHandler(console)
    logging.getLogger(__name__).setLevel(logging.INFO)
    __logger__ = logging.getLogger(__name__)
    __logger__.error = CallCounted(__logger__.error)
    return __logger__


log = get_logger()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import copy
import concurrent.futures
import json
import koji
import os
import time

from ursa_major.logger import log


KOJI_TASK_RESULTS = {
    'timeout': 0,
    'failed': 1,
    'done': 2,
}


def wait_koji_task_finish(koji_session, task_id, timeout, interval):
    """
    Wait for a koji task to finish or stop until timeout.

    :param task_id: koji task id
    :param timeout: timeout in seconds
    :param interval: sleep interval in seconds
    :return: value in KOJI_TASK_RESULTS
    """
    failed_states = [koji.TASK_STATES['CANCELED'], koji.TASK_STATES['FAILED']]
    done_states = [koji.TASK_STATES['CLOSED']]
    unfinish_states = [koji.TASK_STATES['FREE'],
                       koji.TASK_STATES['OPEN'],
                       koji.TASK_STATES['ASSIGNED']]

    start = time.time()
    while True:
        if (time.time() - start) >= timeout:
            return KOJI_TASK_RESULTS['timeout']

        task_info = koji_session.getTaskInfo(task_id)
        state = task_info['state']
        if state in unfinish_states:
            log.info("Waiting for task %s to finish...", task_id)
            time.sleep(interval)
        elif state in failed_states:
            return KOJI_TASK_RESULTS['failed']
        elif state in done_states:
            return KOJI_TASK_RESULTS['done']
        else:
            raise RuntimeError("Task %s is in unknown state %s",
                               task_id, state)


class KojiService(object):
    def __init__(self, profile, dry_run=False):
        self.profile = profile
        self.dry_run = dry_run
        self.koji_module = koji.get_profile_module(profile)
        session_opts = koji.grab_session_options(self.koji_module.config)
        self.koji_proxy = koji.ClientSession(self.koji_module.config.server,
                                             session_opts)

    def login(self):
        """Authenticate to the hub."""
        if self.koji_proxy.logged_in:
            return

        auth_type = self.koji_module.config.authtype

        has_cert = (
            auth_type is None and
            os.path.isfile(os.path.expanduser(self.koji_module.config.cert)))

        if auth_type == 'ssl' or has_cert:
            self.koji_proxy.ssl_login(
                os.path.expanduser(self.koji_module.config.cert),
                os.path.expanduser(self.koji_module.config.ca),
                os.path.expanduser(self.koji_module.config.serverca))

        elif auth_type == 'kerberos':
            self.koji_proxy.gssapi_login(
                getattr(self.koji_module.config, 'principal', None),
                getattr(self.koji_module.config, 'keytab', None))
        else:
            raise RuntimeError('Unsupported authentication type in Koji')

    def logout(self):
        """ expire the login session"""
        if self.koji_proxy.logged_in:
            self.koji_proxy.logout()

    def get_tag(self, tag):
        """Get tag info

        :param tag: tag name or ID.
        :type tag: int or str
        :return: the return value from Koji API getTag.
        """
        log.debug('Get tag info: %s', tag)
        return self.koji_proxy.getTag(tag)

    def get_tag_id(self, tag):
        """Get tag id

        :param tag: tag name.
        :type tag: str
        :return: tag id
        """
        taginfo = self.get_tag(tag)
        return taginfo['id']

    def is_build_tag(self, tag):
        """ Check if tag is a build tag

        :param tag: tag name
        :return: return True if tag is a build tag, else False
        """
        if self.get_build_targets(self.get_tag_id(tag)):
            return True
        return False

    def find_build_tags(self, tag):
        """
        Find build tags in tag's children tags
        It stops when reaches the first build tag in each inheritance path.

        :param tag: tag name
        :return: List of tag names
        """
        # To find out the build tags, we check the tag's inheritance data,
        # when it reaches the first build tag in each inheritant path, it
        # stops and return that build tag, and it stops at any tag that name
        # starts with 'module-'.
        #
        # For examples, if we have tag inheritance data as below (tags with
        # '*' marks are build tags):
        #
        # Example #1:
        #     my-example-tag
        #       └─product-foo-temp-override
        #          └─product-foo-override
        #             └─product-foo-build (*)
        #                ├─tmp-product-foo-build (*)
        #                └─alt-product-foo-build (*)
        #
        # Build tags for 'my-example-tag' are: ['product-foo-build']
        #
        # In example #1, we stops at 'product-foo-build', so
        # 'tmp-product-foo-build' and 'alt-product-foo-build' are not checked
        # at all.
        #
        # Example #2:
        #     my-example-tag
        #       ├─module-345678-build
        #       ├─module-234567-build
        #       ├─module-123456-build
        #       │  └─product-foo-module-hotfix
        #       │     └─product-foo-module-hotfix-build (*)
        #       ├─tmp-product-foo-python-candidate
        #       │  └─tmp-product-foo-python-override
        #       │     └─tmp-product-foo-python-build (*)
        #       ├─product-foo-container-build (*)
        #       └─product-foo-temp-override
        #          └─product-foo-override
        #             └─product-foo-build (*)
        #                ├─tmp-product-foo-build (*)
        #                └─alt-product-foo-build (*)
        #
        # Build tags for 'my-example-tag' are:
        #
        #     ['tmp-product-foo-python-build', 'product-foo-container-build',
        #      'product-foo-build']
        #
        # In example #2, 'product-foo-module-hotfix-build' is a build tag, but
        # we doesn't count it in, because we stops at its child tag
        # 'module-123456-build'.

        tag_id = self.get_tag_id(tag)

        inheritance = self.koji_proxy.getFullInheritance(tag_id, reverse=True)
        # sort the tags by depth, so we can check tags will lower depth first
        # and if a tag match the condition which we will stop, we can add it
        # in blocked tags, then it's children (which has higher depth) will
        # not be checked either.
        sorted_inheritance = sorted(inheritance, key=lambda x: x['currdepth'])

        build_tags = []

        # check whether a tag is build tag, but not check tags that:
        # 1. tag name starts with 'module-'
        # 2. its parent tag is in blocked_tags
        blocked_tags = []

        for tag in sorted_inheritance:
            parent = tag['parent_id']
            if parent in blocked_tags:
                blocked_tags.append(tag['tag_id'])
                continue

            # ignore all 'module-*' tags and don't continue with their children
            if tag['name'].startswith('module-'):
                blocked_tags.append(tag['tag_id'])
                continue

            log.info("Checking whether '%s' is a build tag", tag['name'])
            if self.is_build_tag(tag['name']):
                blocked_tags.append(tag['tag_id'])
                build_tags.append(tag['name'])
        return build_tags

    def get_inheritance_data(self, tag, full=False):
        """Get koji tag inheritance data

        :param tag: tag name or id
        :param full: if true, this will return the full inheritance data
        """
        if full:
            return self.koji_proxy.getFullInheritance(tag)
        return self.koji_proxy.getInheritanceData(tag)

    def set_inheritance_data(self, tag, data):
        """A simple wrapper of Koji API setInheritanceData to make it easy to dry run.

        :param tag: tag id or name.
        :type tag: int or str
        :param data: list of mappings containing tag update information that
            will be done by Koji.
        :type data: list[dict]
        """
        if self.dry_run:
            log.info('DRY RUN: set inheritance data for tag %s: %s',
                     tag, json.dumps(data, indent=2))
        else:
            return self.koji_proxy.setInheritanceData(tag, data)

    def update_tag_inheritance(self, tag, add_tags=None, remove_tags=None):
        """Update tag inheritance data with specified data of tags.

        :param tag:         id or name of child tag
        :param add_tags:    list of parent tags to be added, each tag is a
                            dict containing at least tag name or id
        :param remove_tags: list of tags to be removed, each tag is a dict
                            containing at least tag name or id
        :return: a tuple of (added_tags, removed_tags)
        :rtype: tuple[list[str], list[str]]
        """
        # lists to store the added/updated and removed tags, return them
        # after at the end
        added_tags_list = []
        removed_tags_list = []

        child_tag = self.get_tag(tag)
        if not child_tag:
            raise RuntimeError("Unknown koji tag: {}".format(tag))

        if add_tags is None and remove_tags is None:
            log.info("There is no update data specified for updating tag "
                     "inheritance of '%s', do nothing", tag)
            return

        inheritance_data = self.get_inheritance_data(child_tag['id'])
        log.debug("Inheritance data of tag %s:\n%s",
                  child_tag['name'], json.dumps(inheritance_data, indent=2))
        parent_ids = [d['parent_id'] for d in inheritance_data]

        if remove_tags is not None:
            for tag in remove_tags:
                tag_name_or_id = tag.get('name') or tag.get('id')
                remove_tag = self.get_tag(tag_name_or_id)
                if not remove_tag:
                    raise RuntimeError("Tag {} doesn't exist".format(tag_name_or_id))

                if remove_tag['id'] not in parent_ids:
                    msg = "Tag {} not found in inheritance.".format(tag_name_or_id)
                    log.error(msg)
                    raise RuntimeError(msg)

                data = next(d for d in inheritance_data if d['parent_id'] == remove_tag['id'])

                remove_data = data.copy()
                remove_data['delete link'] = True
                index = inheritance_data.index(data)
                inheritance_data[index] = remove_data
                removed_tags_list.append(remove_data['name'])

        if add_tags is not None:
            # update inheritance data for tags need to be added or updated
            for tag in add_tags:
                add_tag = self.get_tag(tag.get('name') or tag.get('id'))
                if not add_tag:
                    raise RuntimeError("Tag {} doesn't exist".format(tag))

                # construct the inheritance data with all fields in koji so
                # we can compare later if this tag already exists
                add_data = {}
                add_data['child_id'] = child_tag['id']
                add_data['parent_id'] = add_tag.get('id')
                add_data['name'] = add_tag.get('name')
                add_data['priority'] = tag.get('priority', 0)
                add_data['maxdepth'] = tag.get('maxdepth', None)
                add_data['intransitive'] = tag.get('intransitive', False)
                add_data['noconfig'] = tag.get('noconfig', False)
                add_data['pkg_filter'] = tag.get('pkg_filter', '')

                # check priority conflict, filter out the existing same tag
                # and the tags we're going to remove
                same_priority = [d for d in inheritance_data if
                                 d['priority'] == add_data['priority'] and
                                 d['parent_id'] != add_data['parent_id'] and
                                 not d.get('delete link', False)]
                if same_priority:
                    # TODO: should we raise error or just warn
                    msg = ("Tag {} has an inheritance with the same priority "
                           "{} as tag {}".format(child_tag['name'],
                                                 add_data['priority'],
                                                 add_tag['name']))
                    log.warning(msg)

                if add_data['parent_id'] in parent_ids:
                    # this is a tag already exists in inheritance data
                    # we're going to update its data
                    data = next(d for d in inheritance_data
                                if d['parent_id'] == add_data['parent_id'])
                    if data == add_data:
                        log.info("Inheritance %s has same data as existing, skipping.", add_data)
                    else:
                        index = inheritance_data.index(data)
                        inheritance_data[index] = add_data
                        added_tags_list.append(add_data['name'])
                else:
                    inheritance_data.append(add_data)
                    added_tags_list.append(add_data['name'])

        if added_tags_list or removed_tags_list:
            log.info(
                'Update inheritance of tag %s with inheritance data: %s',
                child_tag['name'], json.dumps(inheritance_data, indent=2))
            self.set_inheritance_data(child_tag['id'], inheritance_data)
        return (added_tags_list, removed_tags_list)

    def remove_tags_from_inheritance(self, tag, tags):
        """ Remove tags from a tag's inheritance data.

        :param tag: tag name, the child tag where we get the inheritance data
        :param tags: a list of tag names
        """
        inheritance = self.get_inheritance_data(tag)
        for remove_tag in tags:
            taginfo = self.get_tag(remove_tag)
            data = [d for d in inheritance
                    if d['parent_id'] == taginfo['id']]
            if len(data) == 0:
                raise RuntimeError("Tag {} is not in tag {}'s inheritance"
                                   .format(remove_tag, tag))
            data = data[0]
            remove_data = copy.deepcopy(data)
            remove_data['delete link'] = True
            index = inheritance.index(data)
            inheritance[index] = remove_data

        msg = ("{}Removing tags {!r} from tag {}'s inheritance"
               .format("DRYRUN: " if self.dry_run else "", tags, tag))
        log.info(msg)
        if not self.dry_run:
            self.login()
            self.set_inheritance_data(tag, inheritance)

    def get_build_targets(self, build_tag_id):
        """
        Get build targets of a tag.

        :param build_tag_id: tag id
        :return: return value from koji's getBuildTargets API
        """
        if self.dry_run:
            log.info("DRY RUN: Getting build target for tag %s",
                     build_tag_id)
            return []
        return self.koji_proxy.getBuildTargets(buildTagID=build_tag_id)

    def regen_repo(self, tag, debuginfo=False, src=False, wait=False):
        """
        Regenarate a repo.

        :param tag: tag name or tag id
        :param debuginfo: include debuginfo rpms in repo or not
        :param src: include source rpms in repo or not
        """
        taginfo = self.get_tag(tag)
        if not taginfo:
            msg = "Tag {} not found".format(tag)
            log.error(msg)
            raise RuntimeError(msg)

        tagid = taginfo['id']
        tagname = taginfo['name']
        if not taginfo['arches']:
            log.warning("Warning: tag %s has an empty arch list", tagname)

        targets = self.get_build_targets(tagid)
        if not targets:
            log.warning("Warning: %s is not a build tag", tagname)

        repo_opts = {}
        if debuginfo:
            repo_opts['debuginfo'] = True
        if src:
            repo_opts['src'] = True

        if self.dry_run:
            log.info("DRY RUN: Regenerating repo for tag: %s", tagname)
            return None

        try:
            task_id = self.koji_proxy.newRepo(tagname, **repo_opts)
        except Exception as e:
            raise RuntimeError("Failed to create regenerate repo task: %s" % str(e))
        task_url = "{}/taskinfo?taskID={}".format(
            self.koji_module.config.weburl, task_id)
        log.info("Regenerating repo for tag: %s", tagname)
        log.info("Created task: %d", task_id)
        log.info("Task info: %s", task_url)
        if not wait:
            return

        state = wait_koji_task_finish(self.koji_proxy, task_id,
                                      timeout=30 * 60, interval=10)
        if state == KOJI_TASK_RESULTS['timeout']:
            msg = "Timeout while waiting task {} to finish".format(task_id)
            raise RuntimeError(msg)
        elif state == KOJI_TASK_RESULTS['failed']:
            msg = "Task {} failed, check task url {} for details".format(
                task_id, task_url)
            raise RuntimeError(msg)
        elif state == KOJI_TASK_RESULTS['done']:
            log.info("Task %s succeed, repo of tag %s is re-generated",
                     task_id, tagname)

    def regen_repos(self, tags, debuginfo=False, src=False, wait=False):
        """
        Regenerate repos of tags

        :param tags: list of tag names
        :return: None
        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(tags)) as executor:
            to_regen = {
                executor.submit(self.regen_repo, tag, debuginfo=debuginfo, src=src, wait=wait):
                tag for tag in tags
            }
            for future in concurrent.futures.as_completed(to_regen):
                tag = to_regen[future]
                try:
                    future.result()
                except Exception as exc:
                    log.warning("Error while re-generating repo for %s: %s", tag, str(exc))
                else:
                    if not wait:
                        log.info("Regen-repo task of tag %s is created", tag)
                    else:
                        log.info("Repo of %r is re-generated", tag)

    def list_tags(self, build):
        """
        List tags of a build. Tag inheritance doesn't count.

        :param build: A koji build to query tags
        :type build: int or str or dict
        :return: a list of tags
        :rtype: list[str]
        """
        return [t['name'] for t in self.koji_proxy.listTags(build)]



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import copy
import json
import traceback


from ursa_major import ModuleConfig
from ursa_major.handlers.base import BaseHandler
from ursa_major.logger import log
from ursa_major.mail import MailAPI
from ursa_major.utils import (
    get_env_var,
    match_any,
    mmd_has_requires,
    mmd_has_buildrequires,
    mmd_get_runtime_requires,
    mmd_get_buildtime_requires,
    nvr_to_nsvc,
)


class ModuleInfo(object):
    def __init__(self):
        fields = (
            "build_id",
            "koji_tag",
            "mmd",
            "name",
            "stream",
            "version",
            "context",
            "requires",
            "buildrequires",
            "state",
        )
        for field in fields:
            setattr(self, field, None)

    @classmethod
    def from_mbs_message(cls, mbs, message):
        """
        Create an instance from mbs state change message.

        :param message: message in dict format
        :param mbs: MBS instance
        :return: an instance of ModuleInfo
        """
        return cls._load_modinfo(mbs.get_module_by_id(message["id"]))

    @classmethod
    def from_koji_tag_message(cls, mbs, message):
        """
        Create an instance from koji tag message.

        Notes: mbs API is called anyway, so we set props from api response rather than msg body

        :param message: message in dict format
        :param mbs: MBS instance
        :return: an instance of ModuleInfo
        """
        return cls._load_modinfo(
            mbs.get_module_by_nsvc(
                nvr_to_nsvc(message['build'])
            )
        )

    @classmethod
    def from_ci_message(cls, mbs, message):
        """
        Create an instance from CI message (MBS or Koji).

        The format is not judged here, it's handled in the proxied `from_*_message` classmethods

        :param message: message in dict format
        :param mbs: MBS instance
        :return: an instance of ModuleInfo
        """
        try:
            modinfo = cls.from_mbs_message(mbs, message)
            log.info("MBS stage change message received")
        except KeyError:
            log.debug("message is not a valid MBS state change one: {!s}".format(message))
            try:
                modinfo = cls.from_koji_tag_message(mbs, message)
                log.info("koji tag message received")
            except (KeyError, ValueError):
                log.error("message is not a valid Koji tagging one: {!s}".format(message))
                raise RuntimeError("Invalid message: {!s}".format(message))
        return modinfo

    @classmethod
    def _load_modinfo(cls, moddict):
        modinfo = cls()
        mmd = moddict['mmd']
        modinfo.build_id = moddict['id']
        modinfo.mmd = moddict['mmd']
        modinfo.koji_tag = moddict['koji_tag']
        modinfo.name = moddict['name']
        modinfo.stream = moddict['stream']
        modinfo.version = moddict['version']
        modinfo.context = moddict['context']
        modinfo.state = moddict['state_name']

        modinfo.requires = mmd_get_runtime_requires(mmd)
        modinfo.buildrequires = mmd_get_buildtime_requires(mmd)

        return modinfo

    @property
    def nsvc(self):
        return "{}:{}:{}:{}".format(self.name, self.stream, self.version, self.context)

    @property
    def nvr(self):
        return "{}-{}-{}.{}".format(self.name, self.stream, self.version, self.context)

    @property
    def short_info(self):
        return {
            "build_id": self.build_id,
            "koji_tag": self.koji_tag,
            "nsvc": self.nsvc,
            "requires": self.requires,
        }


class AddTagHandler(BaseHandler):
    def __init__(self, *args, **kwargs):
        super(AddTagHandler, self).__init__(*args, **kwargs)
        self.send_mail = False
        self.wait_regen_repo = False

    @staticmethod
    def get_module_config(module_configs, modinfo):
        """
        Get the module config from configs which matches the module represented
        by the modinfo.

        :param module_configs: a list of module config
        :type module_configs: list[dict]
        :param modinfo: instance of ModuleInfo
        :type modinfo: :class:`ModuleInfo`
        :return: the matched module config. None is returned if no module
            config matches the module metadata.
        :rtype: dict
        :raises RuntimeError: if more than one module configs match the
            specified module metadata.
        """
        matched = []
        for config in module_configs:
            name = config["name"]
            stream = config["stream"]
            if not (name == modinfo.name and stream == modinfo.stream):
                continue
            dep_requires = config.get("buildrequires")
            if dep_requires and not mmd_has_buildrequires(modinfo.mmd, dep_requires):
                continue
            requires = config.get("requires", {})
            if requires and not mmd_has_requires(modinfo.mmd, requires):
                continue
            matched.append(config)

        if len(matched) == 0:
            return None
        if len(matched) > 1:
            error = (
                "Multiple config items match with module <{}>, unable to "
                "determine which item we should use: {!s}".format(modinfo.nsvc, matched)
            )
            log.error(error)
            raise RuntimeError(error)
        return matched[0]

    def add_module_to_tag(self, tag, modinfo, module_config):
        """
        Add koji tag of module represent by modinfo to specified tag's
        inheritance data, config options are in module_config.

        Note: while adding module tag to inheritance, we remove old tags
        of the module as well. We do it in one call to avoid race
        condition.

        :param tag: tag name
        :param modinfo: ModuleInfo instance
        :param module_config: a module config object.
        :type module_config: :class:`ModuleConfig`
        :return: a tuple of (added_tags, removed_tags)
        :rtype: tuple[list[str], list[str]]
        """
        old_tags = self.find_module_tags_from_koji_inheritance(tag, module_config)

        # if this is a tag already exists, we will try to update it
        if modinfo.koji_tag in old_tags:
            old_tags.remove(modinfo.koji_tag)

        add_tag_info = {"name": modinfo.koji_tag, "priority": module_config.priority}

        self.koji.login()
        added_tags, removed_tags = self.koji.update_tag_inheritance(
            tag,
            add_tags=[add_tag_info],
            remove_tags=[{"name": name} for name in old_tags],
        )

        # regen repo for build tags only when inheritance data is updated
        if added_tags or removed_tags:
            if self.koji.is_build_tag(tag):
                build_tags = [tag]
            else:
                build_tags = self.koji.find_build_tags(tag)

            if not build_tags:
                log.warning("No build tag found for %s, not regen any repo", tag)
            else:
                log.info("Found build tags for '%s': %r", tag, build_tags)
                self.koji.regen_repos(build_tags, wait=self.wait_regen_repo)

        return added_tags, removed_tags

    def run(self):
        try:
            ci_message_string = get_env_var(
                self.args.module_from_env, raise_if_not_exist=True
            )
            ci_message = json.loads(ci_message_string)
            log.info(
                "Received MBS state change or Koji tagging message:\n%s",
                json.dumps(ci_message, indent=2, sort_keys=True),
            )

        except Exception as e:
            error = "Fail to load '{!s}' from os environ: {!s}".format(
                self.args.module_from_env, e
            )
            log.error(error)
            raise RuntimeError(error)

        if 'koji_tag' in ci_message and ci_message.get('state_name', None) != 'ready':
            log.info("MBS build state is not 'ready', skipping")
            return

        self.dry_run = self.args.dry_run
        self.send_mail = self.args.send_mail
        self.tag_config_file = self.args.tag_config_file
        self.wait_regen_repo = self.args.wait_regen_repo

        self.load_tag_config()

        if self.send_mail:
            if not self.config.has_section("mail"):
                log.error(
                    "Sending mail is enabled, but config doesn't contain mail settings"
                )
                raise RuntimeError("Missing mail settings")
            mail_config = dict(self.config.items("mail"))
            self.mail_api = MailAPI(mail_config)
            self.mail_data = {"build_url": get_env_var("BUILD_URL")}

        self.connect_koji(dry_run=self.dry_run)
        self.connect_mbs()

        modinfo = ModuleInfo.from_ci_message(self.mbs, ci_message)

        if self.send_mail:
            self.mail_data["module"] = modinfo.short_info

        module_tags = self.koji.list_tags(modinfo.nvr)
        log.info("module build: {} is in tags: {}".format(modinfo.nsvc, module_tags))

        errors = []
        for tag, tag_config in self.tag_config.items():
            tag_owners = tag_config.get("owners", [])
            if not tag_owners:
                log.warning(
                    "Sending mail is enabled, but no owners set for tag %s, "
                    "will not send mail",
                    tag
                )

            conf_modules = tag_config.get("modules", [])

            tags_from = tag_config.get("tags_from", self.default_tags_from)
            if not match_any(module_tags, tags_from):
                log.info(
                    "Tags of module build: {} are not in any of {},"
                    " skipping tag: {}".format(modinfo.nsvc, tags_from, tag)
                )
                continue

            try:
                module_config = self.get_module_config(conf_modules, modinfo)
                # not found in .modules list
                if module_config is None:
                    log.info(
                        "Module %s is not listed in tag %s, skipping this tag.",
                        modinfo.nsvc,
                        tag,
                    )
                    continue

                # the module we're handling matches config of this tag
                added_tags, removed_tags = self.add_module_to_tag(
                    tag, modinfo, ModuleConfig(module_config)
                )

                if self.dry_run or not tag_owners:
                    continue

                if (added_tags or removed_tags) and self.send_mail and tag_owners:
                    # send tag update success mail
                    mail_subject = "Tag {} is updated".format(tag)
                    content = copy.deepcopy(self.mail_data)
                    content["tag"] = tag
                    content["added_tags"] = added_tags
                    content["removed_tags"] = removed_tags
                    log.info(
                        "Sending mail to %r with content:\n%r", str(tag_owners), content
                    )
                    self.mail_api.send_mail(
                        "add-tag.txt", tag_owners, subject=mail_subject, data=content
                    )

            except Exception as e:
                error = "Error while adding module <{!s}> to tag <{}>: " "{!s}".format(
                    modinfo.nsvc, tag, e
                )
                errors.append(error)
                log.error(error)

                if self.dry_run or not tag_owners:
                    continue

                if self.send_mail and tag_owners:
                    mail_subject = "Failed to update tag {}".format(tag)
                    content = copy.deepcopy(self.mail_data)
                    content["tag"] = tag
                    content["traceback"] = traceback.format_exc()
                    log.info("Sending mail to %r with data:\n%r", tag_owners, content)
                    self.mail_api.send_mail(
                        "add-tag-error.txt",
                        tag_owners,
                        subject=mail_subject,
                        data=content,
                    )

        if errors:
            raise RuntimeError("Errors: {!s}".format(errors))



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import json

from ursa_major.koji_service import KojiService
from ursa_major.logger import log
from ursa_major.mbs import MBS
from ursa_major.utils import lru_cache
from ursa_major import MBS_BUILD_STATES


class BaseHandler(object):
    def __init__(self, config, *args, **kwargs):
        self.config = config

        # global options
        self.dry_run = False

        self._mbs = None
        self._koji = None

        self.tag_config_file = None
        # Loaded from config file on-demand
        self._tag_config = None

    def set_args(self, args):
        self.args = args

    @property
    def tag_config(self):
        if self._tag_config is None:
            raise RuntimeError("tag_config is not set, call load_tag_config first")
        return self._tag_config

    @property
    @lru_cache()
    def default_tags_from(self):
        # PY2 compatible
        raw = self.config.get('main', 'default_tags_from') \
            if self.config.has_option('main', 'default_tags_from') else None
        if raw is not None:
            return [
                i.strip() for i in raw.split(',')
            ] if raw.strip() else []
        return None

    def load_tag_config(self):
        """ Load config from tag config file into self.tag_config. """
        with open(self.tag_config_file, 'r') as f:
            log.info("Loading tag configuration from %s", self.tag_config_file)
            self._tag_config = json.load(f)

    def write_tag_config(self, config=None):
        """ Write the tag config file with content.

        :param config: a dict of json file, if None, content is self.tag_config
        """
        if config is not None:
            content = config
        else:
            content = self.tag_config

        if self.dry_run:
            log.info('DRYRUN: write tag config to file: %s',
                     json.dumps(content, indent=2, sort_keys=True))
        else:
            with open(self.tag_config_file, 'w') as f:
                json.dump(content, f, indent=4, sort_keys=True, separators=(',', ': '))

    @property
    def koji(self):
        if self._koji is None:
            raise RuntimeError("koji is not set, call connect_koji first")
        return self._koji

    def connect_koji(self, dry_run=False):
        """Setup koji client session."""
        koji_profile = self.config.get('koji', 'profile')
        log.info("Connecting to Koji with profile: %s", koji_profile)
        self._koji = KojiService(koji_profile, dry_run=dry_run)

    @property
    def mbs(self):
        if self._mbs is None:
            raise RuntimeError("MBS is not set, call connect_mbs first")
        return self._mbs

    def connect_mbs(self):
        """Create a MBS intance."""
        mbs_url = self.config.get('mbs', 'server_url')
        log.info("Connecting to MBS with url: %s", mbs_url)
        self._mbs = MBS(mbs_url)

    def find_module_tags_from_koji_inheritance(self, tag, module_config):
        """
        Find koji tags in specified tag's inheritance data that belongs to
        the module of module_config.

        :param str tag: tag name. This tag's parent tags will be retrieved from
            Koji and from which to find out module(s)' tag(s).
        :param module_config: instance of ModuleConfig, whose module build(s)
            will be queried from MBS and find out which module(s)' koji_tag are
            the parent tag of ``tag``.
        :type module_config: :class:`ModuleConfig`
        :return: list of tag name(s) which are parent tag of ``tag``.
        :rtype: list[str]
        """
        tags_from_koji = set(t['name'] for t in self.koji.get_inheritance_data(tag))
        modules = self.mbs.get_modules(
            name=module_config.name,
            stream=module_config.stream,
            state=[MBS_BUILD_STATES['ready'], MBS_BUILD_STATES['garbage']])
        tags_from_mbs = set(module['koji_tag'] for module in modules)
        return list(set(tags_from_koji) & set(tags_from_mbs))



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import six
import sys
import json

from ursa_major import MBS_BUILD_STATES
from ursa_major.handlers.base import BaseHandler
from ursa_major.logger import log


class CheckConfigHandler(BaseHandler):

    def validate_mandatory_module_config_fields(self, module_config):
        """ Validate a module's mandatory config fields.

        :param module_config: a module config dict
        :return: a generator iterator error messages
        """

        fields = (
            ('name', six.text_type),
            ('stream', six.text_type),
            ('priority', int),
        )

        for field_name, field_type in fields:
            if field_name not in module_config:
                yield ("Missing {} in module config: {}"
                       .format(field_name, module_config))
                continue
            if not isinstance(module_config.get(field_name), field_type):
                yield ("{} should be {} in module config: {}"
                       .format(field_name, field_type, module_config))

    def run(self):
        self.tag_config_file = self.args.tag_config_file
        self.allow_non_ready_module = self.args.allow_non_ready_module

        self.connect_koji()
        self.connect_mbs()

        self.load_tag_config()

        for tag, config in self.tag_config.items():
            log.info("Checking config of tag %s", tag)
            try:
                tag_inheritance = self.koji.get_inheritance_data(tag)
            except Exception as e:
                log.error("Unable to get inheritance data of tag %s: %s", tag, str(e))
                continue

            # to speed it up we don't really check tags in koji
            if 'tags_from' not in config:
                tags_from = self.default_tags_from
                log.warning(
                    "Tag %s: missing tags_from, use default_tags_from %s in global config file",
                    tag, tags_from
                )
            else:
                tags_from = config.get('tags_from', [])
            if tags_from is None:
                log.warning(
                    "Tag %s, tags_from is empty,"
                    " module builds will be added into BuildRoot without tag restriction.",
                    tag
                )

            tag_owners = config.get('owners', None)
            if tag_owners is None or not tag_owners:
                log.warning("Tag %s: missing owners", tag)
            if not isinstance(tag_owners, list):
                log.warning("Tag %s: owners should be a list of emails", tag)

            module_configs = config.get('modules', [])

            if not module_configs:
                log.warning("Tag %s: no module specified", tag)
                continue

            # save modules by priority for checking duplicate priorities
            # exampe: { 100: [module_1, module_2], 200: [module_3]}
            modules_by_priority = {}

            for module_config in module_configs:
                config_string = json.dumps(module_config, indent=2)
                log.debug("Checking config:\n%s", config_string)

                mandatory_field_errors = []
                for error in self.validate_mandatory_module_config_fields(
                        module_config):
                    mandatory_field_errors.append(error)
                    log.error(error)

                if mandatory_field_errors:
                    continue

                if module_config.get('runtime_deps', None) is not None:
                    log.error("'runtime_deps' in module config is deprecated, "
                              "please use 'requires' instead")
                module_priority = module_config.get('priority')
                modules_with_cur_priority = modules_by_priority.setdefault(module_priority, [])
                modules_with_cur_priority.append(module_config.get('name'))
                if len(modules_with_cur_priority) > 1:
                    log.error("More than one module found with same priority '%s' in "
                              "config:\n%s", module_priority, modules_with_cur_priority)

                modules_of_config = self.mbs.get_modules(
                    requires=module_config.get('requires', {}),
                    buildrequires=module_config.get('buildrequires'),
                    name=module_config['name'],
                    stream=module_config['stream'],
                    state=MBS_BUILD_STATES['ready']
                )

                if not (modules_of_config or self.allow_non_ready_module):
                    log.error("Can't find ready modules in MBS with config: %s", module_config)
                    continue

                # checking existing NAME+STREAM modules, don't include criteria
                # of requires or buildrequires, and not limit to ready modules as
                # some old tags may be garbaged
                existing_modules = self.mbs.get_modules(
                    name=module_config['name'],
                    stream=module_config['stream'],
                    state=[MBS_BUILD_STATES['ready'], MBS_BUILD_STATES['garbage']]
                )
                module_tags = [m['koji_tag'] for m in existing_modules]

                non_priorities = [t['priority'] for t in tag_inheritance
                                  if t['name'] not in module_tags]
                if module_config['priority'] in non_priorities:
                    log.error("Module %s has same priority with a tag in tag %s's inheritance",
                              module_config, tag)

        if log.error.counter > 0:
            log.error("Found error(s) in config")
            sys.exit(1)



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import json
import os
import shutil
import six
import tempfile
import unittest
from argparse import Namespace
from mock import patch
from tests import TEST_DATA_DIR, MockMBSBuildsData, make_mmd, dump_mmd
from ursa_major.handlers.add_module import AddModuleHandler
if six.PY3:  # SafeConfigParser == ConfigParser, former deprecated in >= 3.2
    from six.moves.configparser import ConfigParser
else:
    from six.moves.configparser import SafeConfigParser as ConfigParser


class TestAddModuleHandler(unittest.TestCase):
    def setUp(self):
        config_file = os.path.join(TEST_DATA_DIR, 'ursa-major-test.conf')
        config = ConfigParser()
        config.read(config_file)
        self.handler = AddModuleHandler(config)

        self.tmpdir = tempfile.mkdtemp(suffix='_ursa_major_test')
        self.tag_config_file = os.path.join(self.tmpdir, 'default.json')
        with open(self.tag_config_file, 'w') as f:
            json.dump({}, f)

        self.set_args()

        self.koji_get_profile_module_patcher = patch('koji.get_profile_module')
        self.koji_get_profile_module = self.koji_get_profile_module_patcher.start()

        self.koji_session_patcher = patch('koji.ClientSession')
        self.mock_koji_session = self.koji_session_patcher.start().return_value

        self.request_get_patcher = patch('requests.get')
        self.mock_request_get = self.request_get_patcher.start()
        self.mock_request_get.side_effect = MockMBSBuildsData([]).get

    def tearDown(self):
        self.koji_session_patcher.stop()
        self.request_get_patcher.stop()
        self.koji_get_profile_module_patcher.stop()
        try:
            shutil.rmtree(self.tmpdir)
        except:  # noqa
            pass

    def set_args(self, **kwargs):
        kwargs.setdefault("tag_config_file", self.tag_config_file)
        kwargs.setdefault("tag", "f30-build")
        kwargs.setdefault("module_requires", None)
        kwargs.setdefault("module_buildrequires", None)
        kwargs.setdefault("debug", False)
        kwargs.setdefault("dry_run", False)
        args = Namespace()
        for k, v in kwargs.items():
            setattr(args, k, v)
        self.handler.set_args(args)

    def write_tag_config_file(self, content):
        """ Write content to tag config file.

        :param content: content in dict format
        """
        with open(self.tag_config_file, 'w') as f:
            json.dump(content, f, indent=2, sort_keys=True)

    def read_tag_config_file(self):
        with open(self.tag_config_file, 'r') as f:
            return json.load(f)

    def test_add_new_module_with_nonexist_name_stream(self):
        """
        Add a completely new module.
        """
        self.write_tag_config_file({})
        self.set_args(module_name="testmodule", module_stream="master", priority=50)

        self.handler.run()

        expected = [{
            'name': 'testmodule',
            'stream': 'master',
            'priority': 50
        }]

        tag_configs = self.read_tag_config_file()
        assert expected == tag_configs['f30-build']['modules']

    @patch('ursa_major.handlers.add_module.AddModuleHandler.write_tag_config')
    def test_add_new_module_with_conflict_priority_in_config(self, write_tag_config):
        """
        Add a new module but priority conflicts with existing one in config.
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)

        self.set_args(module_name="mariadb", module_stream="10.4", priority=50)
        with self.assertRaises(ValueError) as ctx:
            self.handler.run()

        self.assertIn("Priority 50 is used by other modules in config",
                      str(ctx.exception))
        write_tag_config.assert_not_called()

    @patch('ursa_major.handlers.add_module.AddModuleHandler.write_tag_config')
    def test_add_new_module_with_conflict_priority_in_koji(self, write_tag_config):
        """
        Add a new module but priority conflicts with existing one in tag inheritance.
        """
        self.write_tag_config_file({})
        mock_builds = [
            {
                'koji_tag': 'module-mariadb-10.4-3020190313091759-a5b0195c',
                'modulemd': dump_mmd(make_mmd(name='mariadb', stream='10.4',
                                              version='3020190313091759', context='a5b0195c',
                                              requires={'platform': 'f30'},
                                              buildrequires={'platform': 'f30'})),
                'name': 'mariadb',
                'stream': '10.4',
                'version': '3020190313091759',
                'context': 'a5b0195c'
            },
            {
                'koji_tag': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'modulemd': dump_mmd(make_mmd(name='mariadb', stream='10.4',
                                              version='3020190304180835', context='a5b0195c',
                                              requires={'platform': 'f30'},
                                              buildrequires={'platform': 'f30'})),
                'name': 'mariadb',
                'stream': '10.4',
                'version': '3020190304180835',
                'context': 'a5b0195c'
            },
        ]
        mock_builds_data = MockMBSBuildsData(mock_builds)
        self.mock_request_get.side_effect = mock_builds_data.get

        inheritance_data = [
            {
                'parent_id': 20,
                'name': 'module-ant-1.10-20181122140939-819b5873',
                'priority': 50,
            },
            {
                'parent_id': 101,
                'name': 'module-mariadb-10.4-3020190304180835-a5b0195c',
                'priority': 20,
            },
        ]
        self.mock_koji_session.getInheritanceData.return_value = inheritance_data

        self.set_args(module_name="mariadb", module_stream="10.4", priority=50)
        with self.assertRaises(ValueError) as ctx:
            self.handler.run()

        self.assertIn("Priority 50 is used by other modules in tag f30-build's inheritance",
                      str(ctx.exception))
        write_tag_config.assert_not_called()

    def test_update_module_with_new_priority(self):
        """
        Update an existing module with a new priority value
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)

        self.set_args(module_name="testmodule", module_stream="master", priority=100)
        self.handler.run()

        expected = [{
            'name': 'testmodule',
            'stream': 'master',
            'priority': 100
        }]

        tag_configs = self.read_tag_config_file()
        self.assertEqual(expected, tag_configs['f30-build']['modules'])

    def test_update_module_with_new_priority_and_requires(self):
        """
        Update an existing module with a new priority and requires
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'requires': {'platform': 'rawhide'},
                        'buildrequires': {'platform': 'rawhide'},
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)
        self.set_args(module_name="testmodule", module_stream="master",
                      module_requires=[('platform', 'f30')],
                      module_buildrequires=[('platform', 'f30')],
                      priority=100)

        self.handler.run()

        expected = [{
            'name': 'testmodule',
            'stream': 'master',
            'priority': 100,
            'requires': {'platform': 'f30'},
            'buildrequires': {'platform': 'f30'},
        }]

        tag_configs = self.read_tag_config_file()
        self.assertEqual(expected, tag_configs['f30-build']['modules'])

    def test_update_module_removing_requires_and_new_priority(self):
        """
        Update an existing module with a new priority and remove requires
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'requires': {'platform': 'rawhide'},
                        'buildrequires': {'platform': 'rawhide'},
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)
        self.set_args(module_name="testmodule", module_stream="master", priority=100)

        self.handler.run()

        expected = [{
            'name': 'testmodule',
            'stream': 'master',
            'priority': 100,
        }]

        tag_configs = self.read_tag_config_file()
        self.assertEqual(expected, tag_configs['f30-build']['modules'])

    @patch('ursa_major.handlers.add_module.AddModuleHandler.write_tag_config')
    def test_add_an_existing_module(self, write_tag_config):
        """
        Add a module which already exists in config file with same config.
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'requires': {'platform': 'f30'},
                        'buildrequires': {'platform': 'f30'},
                        'priority': 50,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)
        self.set_args(module_name="testmodule", module_stream="master",
                      module_requires=[('platform', 'f30')],
                      module_buildrequires=[('platform', 'f30')],
                      priority=50)

        self.handler.run()

        expected = [{
            'name': 'testmodule',
            'stream': 'master',
            'priority': 50,
            'requires': {'platform': 'f30'},
            'buildrequires': {'platform': 'f30'},
        }]

        tag_configs = self.read_tag_config_file()
        self.assertEqual(expected, tag_configs['f30-build']['modules'])
        write_tag_config.assert_not_called()

    @patch('ursa_major.handlers.add_module.AddModuleHandler.write_tag_config')
    def test_multiple_name_stream_modules_in_config(self, write_tag_config):
        """
        Test multiple modules for the same name+stream found in config file
        """
        config_content = {
            'f30-build': {
                'modules': [
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'priority': 50,
                    },
                    {
                        'name': 'testmodule',
                        'stream': 'master',
                        'requires': {'platform': 'f30'},
                        'priority': 100,
                    }
                ],
                'owners': ['owner@example.com']
            }
        }
        self.write_tag_config_file(config_content)

        self.set_args(module_name="testmodule", module_stream="master", priority=200)
        with self.assertRaises(RuntimeError) as ctx:
            self.handler.run()

        self.assertIn("Unexpected error, found multiple modules in config file with "
                      "name 'testmodule' and stream 'master'",
                      str(ctx.exception))
        write_tag_config.assert_not_called()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

""" Various utilities """
from fnmatch import fnmatch
from functools import wraps
import os
import re

import gi
gi.require_version('Modulemd', '2.0')    # noqa: E402
from gi.repository import Modulemd    # noqa: E402
from jinja2 import Environment, PackageLoader    # noqa: E402
import koji    # noqa: E402

from .logger import log    # noqa: E402


try:
    from functools import lru_cache
except ImportError:
    def lru_cache():
        def decorator(user_function):
            cache = {}

            @wraps(user_function)
            def wrapper(*args):
                key = tuple(args)
                if key not in cache:
                    cache[key] = user_function(*args)
                return cache[key]
            return wrapper
        return decorator


def validate_mail_address(addr):
    """
    Validate e-mail address format.
    """

    mail_pattern = r"^[_a-z0-9-]+(.[_a-z0-9-]+)*@[a-z0-9-]+(.[a-z0-9-]+)*(.[a-z]{2,4})$"
    match = re.match(mail_pattern, addr)
    if match:
        return True

    log.error("E-mail address <%s> is not valid.", addr)
    return False


def jinja2_env(category):
    env = Environment(
        loader=PackageLoader("ursa_major", "templates/" + category)
    )

    def regex_replace(s, find, replace):
        return re.sub(find, replace, s)
    env.filters['regex_replace'] = regex_replace

    return env


def get_env_var(var, raise_if_not_exist=False):
    """
    Get variable value from os environ

    :param raise_if_not_exist: If true, raise ValueError when variable is
        not present in os environ, otherwise return None.
    """
    if var not in os.environ and raise_if_not_exist:
        raise ValueError("ENV variable '{}' does not exist.".format(var))
    return os.environ.get(var, None)


def load_mmd(yaml, is_file=False):
    """ Create Modulemd.ModuleStreamV2 object from string or file and return. """
    try:
        mmd = None
        # Modulemd.read_packager_* are supported since libmodulemd 2.11,
        # using old functions.
        if is_file:
            # mmd = Modulemd.read_packager_file(yaml)
            mmd = Modulemd.ModuleStream.read_file(yaml, False, None, None)
        else:
            # mmd = Modulemd.read_packager_string(yaml)
            mmd = Modulemd.ModuleStream.read_string(yaml, False, None, None)
        mmd = mmd.upgrade(Modulemd.ModuleStreamVersionEnum.LATEST)
    except Exception:
        error = 'Invalid modulemd: {0}'.format(yaml)
        log.error(error)
        raise

    return mmd


def requires_included(mmd_requires, config_requires):
    """Test if requires defined in config is included in module metadata

    :param mmd_requires: a mapping representing either buildrequires or
        requires, which is returned from function ``get_buildtime_modules`` and
        ``get_buildtime_streams``, or ``get_runtime_streams`` and
        ``get_runtime_modules``.
    :type mmd_requires: dict[str, Modulemd.SimpleSet]
    :param dict config_requires: a mapping representing either buildrequires or
        requires defined in config file. This is what to check if it is
        included in ``mmd_requires``.
    :return: True if all requires inside ``config_requires`` are included in
        module metadata. Otherwise, False is returned.
    :rtype: bool
    """
    for req_name, req_streams in config_requires.items():
        if req_name not in mmd_requires:
            return False

        if not isinstance(req_streams, list):
            req_streams = [req_streams]

        neg_reqs = set(s[1:] for s in req_streams if s.startswith('-'))
        pos_reqs = set(s for s in req_streams if not s.startswith('-'))

        streams = mmd_requires.get(req_name, set())
        if streams & neg_reqs:
            return False
        if pos_reqs and not (streams & pos_reqs):
            return False
    return True


def mmd_has_requires(mmd, requires):
    """
    Check whether a module represent by the mmd has requires.

    :param mmd: Modulemd.ModuleStreamV2 object
    :param requires: dict of requires, example:
        {'platform': 'f28', 'python3': 'master'}
    """
    return requires_included(mmd_get_runtime_requires(mmd), requires)


def mmd_has_buildrequires(mmd, config_buildrequires):
    """
    Check if a module metadata represented by the mmd has buildrequires.

    :param mmd: a module metadata.
    :type mmd: Modulemd.ModuleStreamV2
    :param dict config_buildrequires: a mapping of buildrequires defined in
        config file to match the module metadata, for example:
        ``{'platform': 'f28', 'python3': 'master'}``.
    :return: True if the specified module metadata has the buildrequires
        defined in config.
    :rtype: bool
    """
    return requires_included(mmd_get_buildtime_requires(mmd),
                             config_buildrequires)


def mmd_get_runtime_requires(mmd):
    """
    Check if a module metadata represented by the mmd has requires.

    :param mmd: a module metadata.
    :type mmd: Modulemd.ModuleStreamV2
    :return: a mapping transformed from requires defined in mmd,
        for example: ``{'platform': set(['f28']), 'python3': set(['master'])}``
    :rtype: dict
    """
    deps_list = mmd.get_dependencies()
    mmd_requires = {}
    if not deps_list:
        mmd_requires = {}
    else:
        mmd_requires_modules = deps_list[0].get_runtime_modules()
        for m in mmd_requires_modules:
            mmd_requires[m] = set(deps_list[0].get_runtime_streams(m))
    return mmd_requires


def mmd_get_buildtime_requires(mmd):
    """
    Check if a module metadata represented by the mmd has buildrequires.

    :param mmd: a module metadata.
    :type mmd: Modulemd.ModuleStreamV2
    :return: a mapping transformed from buildrequires defined in mmd,
        for example: ``{'platform': set(['f28']), 'python3': set(['master'])}``
    :rtype: dict
    """
    deps_list = mmd.get_dependencies()
    mmd_requires = {}
    if not deps_list:
        mmd_requires = {}
    else:
        mmd_requires_modules = deps_list[0].get_buildtime_modules()
        for m in mmd_requires_modules:
            mmd_requires[m] = set(deps_list[0].get_buildtime_streams(m))
    return mmd_requires


def nsvc_to_nvr(nsvc, plain_str=True):
    """
    Generate nvr (dict or str) for koji from an nsvc dict

    :param nsvc: N-S-V-C dict
    :type nsvc: dict[str, Any]
    :param plain_str: if True, generate a string as "N-V-R", otherwise a dict, defaults to True
    :type plain_str: bool, optional
    :return: a dict with name,version,release keys or a string as N-V-R
    :rypte: dict[str, str] or str
    """
    nvr_dict = {
        'name': nsvc['name'],
        'version': nsvc['stream'],
        'release': '{version}.{context}'.format(**nsvc)
    }
    if plain_str:
        return koji.buildLabel(nvr_dict)
    else:
        return nvr_dict


def nvr_to_nsvc(nvr, plain_str=True):
    """
    Generate nsvc (dict or str) for MBS from an nvr dict

    :param nvr: N-V-R dict
    :type nvr: dict[str, Any]
    :param plain_str: if True, generate a string as "N:S:V:C", otherwise a dict, defaults to True
    :type plain_str: bool, optional
    :return: a dict with name,stream,version,context keys or a string as N:S:V:C
    :rtype: dict[str, str] or str
    """
    parts = nvr['release'].split('.')
    if len(parts) != 2:
        raise ValueError("Invalid koji nvr for module build: {!s}".format(nvr))
    nsvc_dict = {
        'name': nvr['name'],
        'stream': nvr['version'],
        'version': parts[0],
        'context': parts[1]
    }
    if plain_str:
        return "{name}:{stream}:{version}:{context}".format(**nsvc_dict)
    else:
        return nsvc_dict


def match_any(seq, globs):
    """
    Check if any string in list matches any of the glob patterns


    :param l: list of string
    :type l: list[str]
    :param globs: list of GLOB patterns
    :type globs: list[str]
    :return: True if any matches any; True if globs is None; True if both are [], otherwise False
    :rtype: bool
    """
    if globs is None or seq == globs == []:
        return True
    return any(fnmatch(t, glob) for t in seq for glob in globs)


def find_latest_module(modules, tags_from, koji):
    """
    Get the latest(first) module within tags_from (the tags of the module build on koji matches
    any of pattern in tags_from).

    If tags_from is None, return the 1st item in modules. This is the old behavior without that
    tags_from restriction.
    If tags_from is [], the matched module shouldn't in any tags. (This might be an overkilling
    feature and we should probabliy remove it if it's not needed)

    Notes: modules list is returned by MBS API, so the expected order is from latest to the oldest.

    :param modules: modules from MBS query API
    :type modules: list[dict]
    :param tags_from: a list of glob patterns for module build tags checking
    :type tags_from: list[string] or None
    :param koji: KojiService object
    :type koji: KojiService
    :return: the latest matched item in modules, None if not matched.
    :rtype: dict or None
    """
    if tags_from is None:
        log.info(
            "No tags_from restriction. The latest module is: %s, its koji_tag is %s",
            nsvc_to_nvr(modules[0]), modules[0]['koji_tag']
        )
        return modules[0]
    else:
        log.info("Looking for modules within tags: %s", tags_from)
        for mod in modules:
            modnvr = nsvc_to_nvr(mod)
            log.debug("Validating tags of module: %s", modnvr)
            module_tags = koji.list_tags(modnvr)
            log.debug("Module build: %s is now in tags: %s", modnvr, module_tags)
            if match_any(module_tags, tags_from):
                log.info("Module build: %s (#%s) match", modnvr, mod["id"])
                return mod
            else:
                log.debug(
                    "Tags of module build: %s are not in any of %s, Ignored",
                    modnvr, tags_from
                )
        else:
            log.warning("Module not found in tags: %s", tags_from)
    return None



# -*- coding: utf-8 -*-
# Copyright (c) 2019  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>

import io
import os
import tempfile
import ursa_major.cli
import pytest

from mock import patch


class TestTagConfigFileCLIOption:
    """Test --tag-config-file"""

    @classmethod
    def setup_class(cls):
        # Create a real file to make json.load work
        fd, cls.tag_config_file = tempfile.mkstemp()
        with io.open(fd, 'w', encoding='utf-8', closefd=True) as f:
            f.write(u'{}')

    @classmethod
    def teardown_class(cls):
        os.unlink(cls.tag_config_file)

    @patch.object(ursa_major.cli, 'AddModuleHandler')
    @patch('ursa_major.cli.ConfigParser')
    @patch('os.path.exists', return_value=True)
    def test_use_default_file(self, exists, ConfigParser, AddModuleHandler):
        cli_cmd = [
            'ursa-major', 'add-module', '--name', 'nodejs', '--stream', '10',
            '--priority', '10', '--tag', 'f30-candidate'
        ]

        with patch('sys.argv', new=cli_cmd):
            with patch.object(ursa_major.cli, 'DEFAULT_TAG_CONFIG_FILE',
                              new=self.tag_config_file):
                ursa_major.cli.main()

        handler = AddModuleHandler.return_value
        pos_args, _ = handler.set_args.call_args
        args = pos_args[0]
        assert args.tag_config_file == self.tag_config_file

    @patch.object(ursa_major.cli, 'AddModuleHandler')
    @patch('ursa_major.cli.ConfigParser')
    @patch('os.path.exists', return_value=True)
    def test_default_file_does_not_exist(self, exists, ConfigParser, AddModuleHandler):
        cli_cmd = [
            'ursa-major', 'add-module', '--name', 'nodejs', '--stream', '10',
            '--priority', '10', '--tag', 'f30-candidate'
        ]

        with patch('sys.argv', new=cli_cmd):
            with patch.object(ursa_major.cli, 'DEFAULT_TAG_CONFIG_FILE', new='/tmp/xxxx'):
                with pytest.raises(SystemExit):
                    ursa_major.cli.main()

    @patch.object(ursa_major.cli, 'AddModuleHandler')
    @patch('ursa_major.cli.ConfigParser')
    @patch('os.path.exists', return_value=True)
    def test_specify_a_file(self, exists, ConfigParser, AddModuleHandler):
        cli_cmd = [
            'ursa-major', 'add-module', '--name', 'nodejs', '--stream', '10',
            '--priority', '10', '--tag', 'f30-candidate',
            '--tag-config-file', self.tag_config_file
        ]

        with patch('sys.argv', new=cli_cmd):
            ursa_major.cli.main()

        handler = AddModuleHandler.return_value
        pos_args, _ = handler.set_args.call_args
        args = pos_args[0]
        assert args.tag_config_file == self.tag_config_file

    @patch.object(ursa_major.cli, 'AddModuleHandler')
    @patch('ursa_major.cli.ConfigParser')
    @patch('os.path.exists', return_value=True)
    def test_specified_file_does_not_exist(self, exists, ConfigParser, AddModuleHandler):
        cli_cmd = [
            'ursa-major', 'add-module', '--name', 'nodejs', '--stream', '10',
            '--priority', '10', '--tag', 'f30-candidate',
            '--tag-config-file', '/tmp/xxxxx'
        ]

        with patch('sys.argv', new=cli_cmd):
            with pytest.raises(SystemExit):
                ursa_major.cli.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Chenxiong Qi <cqi@redhat.com>
#            Qixiang Wan <qwan@redhat.com>

import mock
import unittest

from ursa_major.koji_service import KojiService


class KojiServiceTestCase(unittest.TestCase):
    def setUp(self):
        self.koji_profile = mock.Mock()
        with mock.patch('ursa_major.koji_service.koji') as koji:
            koji.gssapi_login = mock.Mock()
            koji.get_profile_module = mock.Mock(
                return_value=mock.Mock(
                    config=mock.Mock(
                        server='koji.example.com',
                        weburl='koji.example.com',
                        authtype='kerberos',
                        krb_rdns=False,
                        cert=''),
                )
            )
            self.koji_profile = koji.get_profile_module.return_value
            self.koji = KojiService('dummy-koji')
            self.koji.koji_proxy.logged_in = False

    def test_gssapi_login(self):
        self.koji.koji_module.config.keytab = 'testkeytab'
        self.koji.koji_module.config.principal = 'testprincipal'
        self.assertEqual(self.koji.koji_module.config.krb_rdns, False)
        self.koji.login()
        self.koji.koji_proxy.gssapi_login.assert_called_with(
            'testprincipal',
            'testkeytab'
        )

    def test_get_inheritance_data(self):
        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(),
            getInheritanceData=mock.Mock(),
        )
        self.koji.get_inheritance_data(1)
        self.koji.koji_proxy.getInheritanceData.assert_called()
        self.koji.koji_proxy.getFullInheritance.assert_not_called()

    def test_get_full_inheritance_data(self):
        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(),
            getInheritanceData=mock.Mock(),
        )
        self.koji.get_inheritance_data(1, full=True)
        self.koji.koji_proxy.getInheritanceData.assert_not_called()
        self.koji.koji_proxy.getFullInheritance.assert_called()

    def test_find_build_tags_with_only_one_child(self):
        getTag_data = {
            'f28-build': {'name': 'f28-build', 'id': 3},
            'module-testmodule': {'name': 'module-testmodule', 'id': 13},
        }
        getBuildTargets_data = {
            3: [{'id': 100, 'name': 'f28-buildroot'}],
            13: [],
        }
        getFullInheritance_data = {
            3: [],
            13: [
                {'currdepth': 1,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-build',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 13,
                 'pkg_filter': '',
                 'priority': 60,
                 'tag_id': 3},
            ],
        }

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(
                side_effect=lambda name: getTag_data.get(name)),
            getBuildTargets=mock.Mock(
                side_effect=lambda buildTagID: getBuildTargets_data.get(buildTagID)),
            getFullInheritance=mock.Mock(
                side_effect=lambda tag_id, reverse: getFullInheritance_data.get(tag_id)),
        )

        build_tags = self.koji.find_build_tags('module-testmodule')
        self.assertEqual(build_tags, ['f28-build'])

    def test_find_build_tags_with_none_result(self):
        getTag_data = {
            'f28-build': {'name': 'f28-build', 'id': 3},
            'module-testmodule': {'name': 'module-testmodule', 'id': 13},
        }
        getBuildTargets_data = {
            3: [],
            13: [],
        }
        getFullInheritance_data = {
            3: [],
            13: [
                {'currdepth': 1,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-build',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 13,
                 'pkg_filter': '',
                 'priority': 60,
                 'tag_id': 3},
            ],
        }

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(
                side_effect=lambda name: getTag_data.get(name)),
            getBuildTargets=mock.Mock(
                side_effect=lambda buildTagID: getBuildTargets_data.get(buildTagID)),
            getFullInheritance=mock.Mock(
                side_effect=lambda tag_id, reverse: getFullInheritance_data.get(tag_id)),
        )

        build_tags = self.koji.find_build_tags('module-testmodule')
        self.assertEqual(build_tags, [])

    def test_find_build_tags_in_different_depths(self):
        getTag_data = {
            'f28-build': {'name': 'f28-build', 'id': 3},
            'f28-other-build': {'name': 'f28-other-build', 'id': 4},
            'f28-other-build2': {'name': 'f28-other-build2', 'id': 5},
            'f28-non-build': {'name': 'f28-non-build', 'id': 6},
            'f28-other-build3': {'name': 'f28-other-build3', 'id': 7},
            'module-testmodule': {'name': 'module-testmodule', 'id': 13},
        }
        getBuildTargets_data = {
            3: [{'id': 100, 'name': 'f28-buildroot'}],
            4: [{'id': 200, 'name': 'f28-other-buildroot'}],
            5: [{'id': 300, 'name': 'f28-other-buildroot2'}],
            6: [],
            7: [{'id': 400, 'name': 'f28-other-buildroot3'}],
            13: [],
        }
        getFullInheritance_data = {
            3: [],
            13: [
                {'currdepth': 1,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-build',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 13,
                 'pkg_filter': '',
                 'priority': 60,
                 'tag_id': 3},
                {'currdepth': 1,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-other-build',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 13,
                 'pkg_filter': '',
                 'priority': 80,
                 'tag_id': 4},
                {'currdepth': 1,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-non-build',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 13,
                 'pkg_filter': '',
                 'priority': 90,
                 'tag_id': 6},
                {'currdepth': 2,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-other-build2',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 4,
                 'pkg_filter': '',
                 'priority': 100,
                 'tag_id': 5},
                {'currdepth': 2,
                 'filter': [],
                 'intransitive': False,
                 'maxdepth': None,
                 'name': 'f28-other-build3',
                 'nextdepth': None,
                 'noconfig': False,
                 'parent_id': 6,
                 'pkg_filter': '',
                 'priority': 100,
                 'tag_id': 7},
            ],
        }

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(
                side_effect=lambda name: getTag_data.get(name)),
            getBuildTargets=mock.Mock(
                side_effect=lambda buildTagID: getBuildTargets_data.get(buildTagID)),
            getFullInheritance=mock.Mock(
                side_effect=lambda tag_id, reverse: getFullInheritance_data.get(tag_id)),
        )

        build_tags = self.koji.find_build_tags('module-testmodule')
        self.assertEqual(build_tags, ['f28-build', 'f28-other-build', 'f28-other-build3'])

    def test_update_tag_inheritance_successfully(self):
        getTag_data = {
            'f27-build': {'name': 'f27-build', 'id': 3},
            'module-100': {'name': 'module-100', 'id': 13},
            'module-123': {'name': 'module-123', 'id': 1000},
        }
        inheritance_data = [{'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-buildrequires',
                             'noconfig': False,
                             'parent_id': 4,
                             'pkg_filter': '',
                             'priority': 10},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-go-toolset-candidate',
                             'noconfig': False,
                             'parent_id': 12,
                             'pkg_filter': '',
                             'priority': 30},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'module-100',
                             'noconfig': False,
                             'parent_id': 13,
                             'pkg_filter': '',
                             'priority': 40}]

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getInheritanceData=mock.Mock(return_value=inheritance_data),
        )

        add_tags = [{'name': 'module-123', 'priority': 50}]
        remove_tags = [{'name': 'module-100'}]
        self.koji.update_tag_inheritance('f27-build', add_tags, remove_tags)
        module_123 = {'pkg_filter': False,
                      'intransitive': False,
                      'priority': 50,
                      'parent_id': 1000,
                      'maxdepth': None,
                      'noconfig': False}
        inheritance_data.append(module_123)
        inheritance_data[2]['delete link'] = True
        self.koji.koji_proxy.setInheritanceData.assert_called_with(
            3, inheritance_data)

    @mock.patch('ursa_major.koji_service.log')
    def test_update_tag_inheritance_with_same_priority(self, log):
        getTag_data = {
            'f27-build': {'name': 'f27-build', 'id': 3},
            'module-123': {'name': 'module-123', 'id': 1000},
        }
        inheritance_data = [{'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-buildrequires',
                             'noconfig': False,
                             'parent_id': 4,
                             'pkg_filter': '',
                             'priority': 10},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-go-toolset-candidate',
                             'noconfig': False,
                             'parent_id': 12,
                             'pkg_filter': '',
                             'priority': 30}]

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getInheritanceData=mock.Mock(return_value=inheritance_data),
        )

        add_tags = [{'name': 'module-123', 'priority': 30}]
        self.koji.update_tag_inheritance('f27-build', add_tags)

        log.warning.assert_called_with(
            ("Tag f27-build has an inheritance with the same priority 30 as "
             "tag module-123"))

        module_123 = {'pkg_filter': False,
                      'intransitive': False,
                      'priority': 50,
                      'parent_id': 1000,
                      'maxdepth': None,
                      'noconfig': False}
        inheritance_data.append(module_123)
        self.koji.koji_proxy.setInheritanceData.assert_called_with(
            3, inheritance_data)

    @mock.patch('ursa_major.koji_service.log')
    def test_update_tag_inheritance_with_same_tag_name(self, log):
        getTag_data = {
            'f27-build': {'name': 'f27-build', 'id': 3},
            'f27-buildrequires': {'name': 'f27-buildrequires', 'id': 4},
        }
        inheritance_data = [{'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-buildrequires',
                             'noconfig': False,
                             'parent_id': 4,
                             'pkg_filter': '',
                             'priority': 10},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-go-toolset-candidate',
                             'noconfig': False,
                             'parent_id': 12,
                             'pkg_filter': '',
                             'priority': 30}]

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getInheritanceData=mock.Mock(return_value=inheritance_data),
        )

        add_tags = [{'name': 'f27-buildrequires', 'priority': 50}]
        added_tags, removed_tags = self.koji.update_tag_inheritance('f27-build', add_tags)

        self.koji.koji_proxy.setInheritanceData.assert_called_with(
            3,
            [
                {'intransitive': False,
                 'name': 'f27-buildrequires',
                 'pkg_filter': '',
                 'priority': 50,
                 'parent_id': 4,
                 'maxdepth': None,
                 'noconfig': False,
                 'child_id': 3},
                {'intransitive': False,
                 'parent_id': 12,
                 'maxdepth': None,
                 'noconfig': False,
                 'child_id': 3,
                 'pkg_filter': '',
                 'priority': 30,
                 'name': 'f27-go-toolset-candidate'}
            ]
        )
        self.assertEqual(added_tags, ['f27-buildrequires'])
        self.assertEqual(removed_tags, [])

    @mock.patch('ursa_major.koji_service.log')
    def test_update_tag_inheritance_with_removing_non_exist_tag(self, log):
        getTag_data = {
            'f27-build': {'name': 'f27-build', 'id': 3},
            'module-123': {'name': 'module-123', 'id': 1000},
        }
        inheritance_data = [{'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-buildrequires',
                             'noconfig': False,
                             'parent_id': 4,
                             'pkg_filter': '',
                             'priority': 10},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-go-toolset-candidate',
                             'noconfig': False,
                             'parent_id': 12,
                             'pkg_filter': '',
                             'priority': 30}]

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getInheritanceData=mock.Mock(return_value=inheritance_data),
        )

        add_tags = [{'name': 'module-123', 'priority': 50}]
        remove_tags = [{'name': 'non-exist-tag'}]

        with self.assertRaises(RuntimeError) as ctx:
            self.koji.update_tag_inheritance('f27-build', add_tags, remove_tags)

        self.assertIn("Tag non-exist-tag doesn't exist",
                      str(ctx.exception))
        self.koji.koji_proxy.setInheritanceData.assert_not_called()

    @mock.patch('ursa_major.koji_service.log')
    def test_update_tag_inheritance_with_removing_non_parent_tag(self, log):
        getTag_data = {
            'f27-build': {'name': 'f27-build', 'id': 3},
            'module-100': {'name': 'module-100', 'id': 13},
            'module-123': {'name': 'module-123', 'id': 1000},
        }
        inheritance_data = [{'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-buildrequires',
                             'noconfig': False,
                             'parent_id': 4,
                             'pkg_filter': '',
                             'priority': 10},
                            {'child_id': 3,
                             'intransitive': False,
                             'maxdepth': None,
                             'name': 'f27-go-toolset-candidate',
                             'noconfig': False,
                             'parent_id': 12,
                             'pkg_filter': '',
                             'priority': 30}]

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getInheritanceData=mock.Mock(return_value=inheritance_data),
        )

        add_tags = [{'name': 'module-123', 'priority': 50}]
        remove_tags = [{'name': 'module-100'}]

        with self.assertRaises(RuntimeError) as ctx:
            self.koji.update_tag_inheritance('f27-build', add_tags, remove_tags)

        self.assertIn("Tag module-100 not found in inheritance.",
                      str(ctx.exception))
        self.koji.koji_proxy.setInheritanceData.assert_not_called()

    @mock.patch('ursa_major.koji_service.wait_koji_task_finish')
    @mock.patch('ursa_major.koji_service.log')
    def test_regen_repo_without_wait_finish(self, log, wait_koji):
        getTag_data = {
            'rhel-8.0-build': {
                'name': 'rhel-8.0-build',
                'id': 100,
                'arches': [],
            },
        }

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getBuildTargets=mock.Mock(return_value=[]),
            newRepo=mock.Mock(return_value=123),
        )

        self.koji.regen_repo('rhel-8.0-build', debuginfo=True, src=True)

        self.koji.koji_proxy.newRepo.assert_called_with(
            'rhel-8.0-build', debuginfo=True, src=True)
        log.warning.assert_has_calls(
            [mock.call('Warning: %s is not a build tag', 'rhel-8.0-build'),
             mock.call('Warning: tag %s has an empty arch list',
                       'rhel-8.0-build')],
            any_order=True)
        log.info.assert_has_calls(
            [mock.call('Regenerating repo for tag: %s', 'rhel-8.0-build'),
             mock.call('Created task: %d', 123),
             mock.call('Task info: %s',
                       'koji.example.com/taskinfo?taskID=123')],
            any_order=True)
        wait_koji.assert_not_called()

    @mock.patch('ursa_major.koji_service.log')
    def test_regen_repo_with_wait_finish(self, log):
        getTag_data = {
            'rhel-8.0-build': {
                'name': 'rhel-8.0-build',
                'id': 100,
                'arches': [],
            },
        }

        self.koji.koji_proxy = mock.Mock(
            getTag=mock.Mock(side_effect=lambda name: getTag_data.get(name)),
            getBuildTargets=mock.Mock(return_value=[]),
            newRepo=mock.Mock(return_value=123),
            getTaskInfo=mock.Mock(return_value={'state': 2}),
        )

        self.koji.regen_repo('rhel-8.0-build',
                             debuginfo=True, src=True, wait=True)

        self.koji.koji_proxy.newRepo.assert_called_with(
            'rhel-8.0-build', debuginfo=True, src=True)
        log.warning.assert_has_calls(
            [mock.call('Warning: %s is not a build tag', 'rhel-8.0-build'),
             mock.call('Warning: tag %s has an empty arch list',
                       'rhel-8.0-build')],
            any_order=True)
        self.koji.koji_proxy.getTaskInfo.assert_called_with(123)

    def test_list_tags(self):
        listTags_data = [{'name': 'foo'}, {'name': 'bar'}]
        self.koji.koji_proxy = mock.Mock(
            listTags=mock.Mock(return_value=listTags_data),
        )
        tags = self.koji.list_tags(1)
        self.koji.koji_proxy.listTags.assert_called_once_with(1)
        self.assertEqual(tags, ['foo', 'bar'])
